@@ -1,31 +1,21 @@
-/* npth.c - a lightweight implementation of pth over pthread.
-   Copyright (C) 2011, 2014 g10 Code GmbH
-
-   This file is part of NPTH.
-
-   NPTH is free software; you can redistribute it and/or modify it
-   under the terms of either
-
-   - the GNU Lesser General Public License as published by the Free
-   Software Foundation; either version 3 of the License, or (at
-   your option) any later version.
-
-   or
-
-   - the GNU General Public License as published by the Free
-   Software Foundation; either version 2 of the License, or (at
-   your option) any later version.
-
-   or both in parallel, as here.
-
-   NPTH is distributed in the hope that it will be useful, but WITHOUT
-   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
-   License for more details.
-
-   You should have received a copies of the GNU General Public License
-   and the GNU Lesser General Public License along with this program;
-   if not, see <http://www.gnu.org/licenses/>.  */
+/* npth.c - a lightweight implementation of pth over native threads
+ * Copyright (C) 2011, 2014 g10 Code GmbH
+ *
+ * This file is part of nPth.
+ *
+ * nPth is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU Lesser General Public License as
+ * published by the Free Software Foundation; either version 2.1 of
+ * the License, or (at your option) any later version.
+ *
+ * nPth is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
+ * the GNU Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this program; if not, see <https://www.gnu.org/licenses/>.
+ */
 
 /* We implement the join mechanism ourself.  */
 
@@ -35,12 +25,13 @@
 
 #include <assert.h>
 #include <errno.h>
+#include <io.h>
 
 #include "npth.h"
 
 #include <stdio.h>
 #define DEBUG_CALLS 1
-#define _npth_debug(x, ...) printf(__VA_ARGS__)
+#define _npth_debug(x, ...) fprintf(stderr, __VA_ARGS__)
 
 #ifndef TEST
 #undef  DEBUG_CALLS
@@ -54,8 +45,20 @@
 
 /* The global lock that excludes all threads but one.  Note that this
    implements the single-user-thread policy, but also protects all our
-   global data such as the thread_table.  */
+   global data such as the thread_table.  GOT_SCEPTRE is a flag used
+   for debugging to tell wether we hold SCEPTRE.  */
 static CRITICAL_SECTION sceptre;
+static int got_sceptre;
+
+
+/* This flag is set as soon as npth_init has been called or if any
+ * thread has been created.  It will never be cleared again.  The only
+ * purpose is to make npth_protect and npth_unprotect more robust in
+ * that they can be shortcut when npth_init has not yet been called.
+ * This is important for libraries which want to support nPth by using
+ * those two functions but may have been initialized before nPth. */
+static int initialized_or_any_threads;
+
 
 typedef struct npth_impl_s *npth_impl_t;
 #define MAX_THREADS 1024
@@ -150,8 +153,9 @@ enter_npth (const char *function)
   int res;
 
   if (DEBUG_CALLS)
-    _npth_debug (DEBUG_CALLS, "enter_npth (%s)\n",
-		 function ? function : "unknown");
+    _npth_debug (DEBUG_CALLS, "tid %lu: enter_npth (%s)\n",
+		 npth_self (), function ? function : "unknown");
+  got_sceptre = 0;
   LeaveCriticalSection (&sceptre);
 }
 
@@ -160,10 +164,11 @@ static void
 leave_npth (const char *function)
 {
   EnterCriticalSection (&sceptre);
+  got_sceptre = 1;
 
   if (DEBUG_CALLS)
-    _npth_debug (DEBUG_CALLS, "leave_npth (%s)\n",
-		 function ? function : "");
+    _npth_debug (DEBUG_CALLS, "tid %lu: leave_npth (%s)\n",
+		 npth_self (), function ? function : "");
 }
 
 #define ENTER() enter_npth(__FUNCTION__)
@@ -320,6 +325,9 @@ npth_init (void)
 
   InitializeCriticalSection (&sceptre);
 
+  /* Track that we have been initialized.  */
+  initialized_or_any_threads = 1;
+
   /* Fake a thread table item for the main thread.  */
   tls_index = TlsAlloc();
   if (tls_index == TLS_OUT_OF_INDEXES)
@@ -1745,16 +1753,33 @@ npth_sendmsg (int fd, const struct msghdr *msg, int flags)
 void
 npth_unprotect (void)
 {
-  ENTER();
+  /* If we are not initialized we may not access the semaphore and
+   * thus we shortcut it. Note that in this case the unprotect/protect
+   * is not needed.  For failsafe reasons if an nPth thread has ever
+   * been created but nPth has accidentally not initialized we do not
+   * shortcut so that a stack backtrace (due to the access of the
+   * uninitialized semaphore) is more expressive.  */
+  if (initialized_or_any_threads)
+    ENTER();
 }
 
 
 void
 npth_protect (void)
 {
-  LEAVE();
+  /* See npth_unprotect for commentary.  */
+  if (initialized_or_any_threads)
+    LEAVE();
 }
 
+
+int
+npth_is_protected (void)
+{
+  return got_sceptre;
+}
+
+
 
 /* Maximum number of extra handles.  We can only support 31 as that is
    the number of bits we can return.  This is smaller than the maximum
